@@ -5,13 +5,13 @@
 Defines the base classes, types, and interface protocols used by
 graphicle's modules.
 """
+import collections.abc as cla
+import typing as ty
 from abc import ABC, abstractmethod
-from typing import Union, Any, Optional, Protocol
 
 import numpy as np
 import numpy.typing as npt
 
-
 __all__ = [
     "DoubleVector",
     "ComplexVector",
@@ -20,6 +20,8 @@ __all__ = [
     "HalfIntVector",
     "ObjVector",
     "AnyVector",
+    "VoidVector",
+    "DoubleUfunc",
     "EventInterface",
     "ArrayBase",
     "ParticleBase",
@@ -34,11 +36,14 @@ BoolVector = npt.NDArray[np.bool_]
 IntVector = npt.NDArray[np.int32]
 HalfIntVector = npt.NDArray[np.int16]
 ObjVector = npt.NDArray[np.object_]
-AnyVector = npt.NDArray[Any]
-MaskLike = Union["MaskBase", BoolVector]
+AnyVector = npt.NDArray[ty.Any]
+VoidVector = npt.NDArray[np.void]
+MaskLike = ty.Union["MaskBase", BoolVector]
+DoubleUfunc = ty.TypeVar("DoubleUfunc", DoubleVector, np.float64)
+DataType = ty.TypeVar("DataType")
 
 
-class EventInterface(Protocol):
+class EventInterface(ty.Protocol):
     """Defines the interface for a generic event object expected by
     graphicle's routines. Attributes are stored as numpy arrays, with
     each element corresponding to a particle in an event. Attributes
@@ -94,23 +99,41 @@ class EventInterface(Protocol):
         ...
 
 
-class ArrayBase(ABC):
+class ArrayBase(ABC, cla.Sequence, np.lib.mixins.NDArrayOperatorsMixin):
     @abstractmethod
-    def __init__(self, data: Optional[AnyVector] = None) -> None:
+    def __init__(self, data: ty.Optional[AnyVector] = None) -> None:
         pass
 
     @abstractmethod
-    def __len__(self) -> int:
-        """Number of elements in array."""
+    def __iter__(self) -> ty.Iterator[ty.Any]:
+        """Iterator exposing contained data as Python native."""
 
     @abstractmethod
     def __bool__(self) -> bool:
         """Truthy returns ``False`` if no elements, ``True`` otherwise."""
 
     @abstractmethod
-    def __array__(self) -> npt.NDArray[Any]:
+    def __array__(self) -> AnyVector:
         """Numpy array representation of the data."""
 
+    @property
+    @abstractmethod
+    def dtype(self) -> AnyVector:
+        """Numpy array representation of the data."""
+
+    @abstractmethod
+    def __eq__(self) -> "MaskBase":
+        """Equality comparison."""
+
+    @abstractmethod
+    def __ne__(self) -> "MaskBase":
+        """Non equality comparison."""
+
+    @property
+    @abstractmethod
+    def _data(self) -> AnyVector:
+        pass
+
     @property
     @abstractmethod
     def data(self) -> AnyVector:
@@ -155,7 +178,7 @@ class ParticleBase(ABC):
         pass
 
 
-class AdjacencyBase(ABC):
+class AdjacencyBase(ABC, cla.Sequence):
     @abstractmethod
     def __len__(self) -> int:
         pass
@@ -197,7 +220,7 @@ class MaskBase(ABC):
         pass
 
     @abstractmethod
-    def __array__(self) -> npt.NDArray[Any]:
+    def __array__(self) -> npt.NDArray[ty.Any]:
         """Numpy array representation of the data."""
 
     @abstractmethod
